# -*- encoding: utf-8 -*-
# Software Management Providers
#
# Copyright (C) 2012 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
#

"""Python Provider for LMI_SoftwareInstalledPackage

Instruments the CIM class LMI_SoftwareInstalledPackage

"""

import pywbem
from pywbem.cim_provider2 import CIMProvider2

from openlmi.common import cmpi_logging
from openlmi.software.core import (
        ComputerSystem, SoftwareFileCheck,
        SoftwareInstalledPackage, SoftwarePackage)
from openlmi.software.yumdb import YumDB

class LMI_SoftwareInstalledPackage(CIMProvider2):
    """Instrument the CIM class LMI_SoftwareInstalledPackage

    The InstalledSoftwareElement association allows the identification of
    the ComputerSystem on which a particular SoftwareElement is installed.

    """

    def __init__ (self, _env):
        cmpi_logging.logger.debug('Initializing provider %s from %s' \
                % (self.__class__.__name__, __file__))

    @cmpi_logging.trace_method
    def get_instance(self, env, model):
        """Return an instance.

        Keyword arguments:
        env -- Provider Environment (pycimmb.ProviderEnvironment)
        model -- A template of the pywbem.CIMInstance to be returned.  The
            key properties are set on this instance to correspond to the
            instanceName that was requested.  The properties of the model
            are already filtered according to the PropertyList from the
            request.  Only properties present in the model need to be
            given values.  If you prefer, you can set all of the
            values, and the instance will be filtered for you.

        Possible Errors:
        CIM_ERR_ACCESS_DENIED
        CIM_ERR_INVALID_PARAMETER (including missing, duplicate, unrecognized
            or otherwise incorrect parameters)
        CIM_ERR_NOT_FOUND (the CIM Class does exist, but the requested CIM
            Instance does not exist in the specified namespace)
        CIM_ERR_FAILED (some other unspecified error occurred)

        """
        if not "Software" in model:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing Software property.")
        if not "System" in model:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing System property.")
        ComputerSystem.check_path_property(env, model, 'System')
        model['System'] = ComputerSystem.get_path()
        with YumDB.getInstance():
            pkg_info = SoftwarePackage.object_path2pkg(model['Software'],
                    kind="installed")
            model['Software'] = SoftwarePackage.pkg2model(
                    pkg_info, keys_only=True)
        return model

    @cmpi_logging.trace_method
    def enum_instances(self, env, model, keys_only):
        """Enumerate instances.

        The WBEM operations EnumerateInstances and EnumerateInstanceNames
        are both mapped to this method.
        This method is a python generator

        Keyword arguments:
        env -- Provider Environment (pycimmb.ProviderEnvironment)
        model -- A template of the pywbem.CIMInstances to be generated.
            The properties of the model are already filtered according to
            the PropertyList from the request.  Only properties present in
            the model need to be given values.  If you prefer, you can
            always set all of the values, and the instance will be filtered
            for you.
        keys_only -- A boolean.  True if only the key properties should be
            set on the generated instances.

        Possible Errors:
        CIM_ERR_FAILED (some other unspecified error occurred)

        """
        # Prime model.path with knowledge of the keys, so key values on
        # the CIMInstanceName (model.path) will automatically be set when
        # we set property values on the model.
        model.path.update({'System': None, 'Software': None})

        yum_package_path = pywbem.CIMInstanceName("LMI_SoftwarePackage",
                namespace=model.path.namespace,
                host=model.path.host)
        model['System'] = ComputerSystem.get_path()
        with YumDB.getInstance() as ydb:
            pkglist = ydb.get_package_list('installed')
            for pkg in pkglist:
                iname = SoftwarePackage.pkg2model(pkg, model=yum_package_path)
                model['Software'] = iname
                yield model

    @cmpi_logging.trace_method
    def set_instance(self, env, instance, modify_existing):
        """Return a newly created or modified instance.

        Keyword arguments:
        env -- Provider Environment (pycimmb.ProviderEnvironment)
        instance -- The new pywbem.CIMInstance.  If modifying an existing
            instance, the properties on this instance have been filtered by
            the PropertyList from the request.
        modify_existing -- True if ModifyInstance, False if CreateInstance

        Return the new instance.  The keys must be set on the new instance.

        Possible Errors:
        CIM_ERR_ACCESS_DENIED
        CIM_ERR_NOT_SUPPORTED
        CIM_ERR_INVALID_PARAMETER (including missing, duplicate, unrecognized
            or otherwise incorrect parameters)
        CIM_ERR_ALREADY_EXISTS (the CIM Instance already exists -- only
            valid if modify_existing is False, indicating that the operation
            was CreateInstance)
        CIM_ERR_NOT_FOUND (the CIM Instance does not exist -- only valid
            if modify_existing is True, indicating that the operation
            was ModifyInstance)
        CIM_ERR_FAILED (some other unspecified error occurred)

        """
        # parse and check arguments
        if modify_existing is True:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_SUPPORTED,
                "MofifyInstance is not supported")

        if not "Software" in instance:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing Software property.")
        if not "System" in instance:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing System property.")
        ComputerSystem.check_path_property(env, instance, 'System')

        with YumDB.getInstance() as ydb:
            pkg_info = SoftwarePackage.object_path2pkg_search(
                    instance['Software'])

            if pkg_info.installed:
                raise pywbem.CIMError(pywbem.CIM_ERR_ALREADY_EXISTS,
                        "Package %s is already installed." % pkg_info)

            cmpi_logging.logger.info('installing package %s' % pkg_info)
            installed_pkg = ydb.install_package(pkg_info)
            cmpi_logging.logger.info('package %s installed' % pkg_info)

        instance["Software"] = SoftwarePackage.pkg2model(installed_pkg)
        return instance

    @cmpi_logging.trace_method
    def delete_instance(self, env, instance_name):
        """Delete an instance.

        Keyword arguments:
        env -- Provider Environment (pycimmb.ProviderEnvironment)
        instance_name -- A pywbem.CIMInstanceName specifying the instance
            to delete.

        Possible Errors:
        CIM_ERR_ACCESS_DENIED
        CIM_ERR_NOT_SUPPORTED
        CIM_ERR_INVALID_NAMESPACE
        CIM_ERR_INVALID_PARAMETER (including missing, duplicate, unrecognized
            or otherwise incorrect parameters)
        CIM_ERR_INVALID_CLASS (the CIM Class does not exist in the specified
            namespace)
        CIM_ERR_NOT_FOUND (the CIM Class does exist, but the requested CIM
            Instance does not exist in the specified namespace)
        CIM_ERR_FAILED (some other unspecified error occurred)

        """
        if not "Software" in instance_name:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing Software property.")
        if not "System" in instance_name:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing System property.")
        ComputerSystem.check_path_property(env, instance_name, 'System')

        with YumDB.getInstance() as ydb:
            pkg_info = SoftwarePackage.object_path2pkg(
                    instance_name["Software"], kind="installed")
            cmpi_logging.logger.info('removing package "%s"' % pkg_info)
            ydb.remove_package(pkg_info)
            cmpi_logging.logger.info('package "%s" removed' % pkg_info)

    @cmpi_logging.trace_method
    def references(self, env, object_name, model, result_class_name, role,
                   result_role, keys_only):
        """Instrument Associations.

        All four association-related operations (Associators, AssociatorNames,
        References, ReferenceNames) are mapped to this method.
        This method is a python generator

        Keyword arguments:
        env -- Provider Environment (pycimmb.ProviderEnvironment)
        object_name -- A pywbem.CIMInstanceName that defines the source
            CIM Object whose associated Objects are to be returned.
        model -- A template pywbem.CIMInstance to serve as a model
            of the objects to be returned.  Only properties present on this
            model need to be set.
        result_class_name -- If not empty, this string acts as a filter on
            the returned set of Instances by mandating that each returned
            Instances MUST represent an association between object_name
            and an Instance of a Class whose name matches this parameter
            or a subclass.
        role -- If not empty, MUST be a valid Property name. It acts as a
            filter on the returned set of Instances by mandating that each
            returned Instance MUST refer to object_name via a Property
            whose name matches the value of this parameter.
        result_role -- If not empty, MUST be a valid Property name. It acts
            as a filter on the returned set of Instances by mandating that
            each returned Instance MUST represent associations of
            object_name to other Instances, where the other Instances play
            the specified result_role in the association (i.e. the
            name of the Property in the Association Class that refers to
            the Object related to object_name MUST match the value of this
            parameter).
        keys_only -- A boolean.  True if only the key properties should be
            set on the generated instances.

        The following diagram may be helpful in understanding the role,
        result_role, and result_class_name parameters.
        +------------------------+                    +-------------------+
        | object_name.classname  |                    | result_class_name |
        | ~~~~~~~~~~~~~~~~~~~~~  |                    | ~~~~~~~~~~~~~~~~~ |
        +------------------------+                    +-------------------+
           |              +-----------------------------------+      |
           |              |  [Association] model.classname    |      |
           | object_name  |  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    |      |
           +--------------+ object_name.classname REF role    |      |
        (CIMInstanceName) | result_class_name REF result_role +------+
                          |                                   |(CIMInstanceName)
                          +-----------------------------------+

        Possible Errors:
        CIM_ERR_ACCESS_DENIED
        CIM_ERR_NOT_SUPPORTED
        CIM_ERR_INVALID_NAMESPACE
        CIM_ERR_INVALID_PARAMETER (including missing, duplicate, unrecognized
            or otherwise incorrect parameters)
        CIM_ERR_FAILED (some other unspecified error occurred)

        """
        cimhandle = env.get_cimom_handle()

        # If you want to get references for free, implemented in terms
        # of enum_instances, just leave the code below unaltered.
        if cimhandle.is_subclass(object_name.namespace,
                          sub=object_name.classname,
                          super='CIM_ComputerSystem') or \
                cimhandle.is_subclass(object_name.namespace,
                               sub=object_name.classname,
                               super='LMI_SoftwarePackage'):
            return self.simple_refs(env, object_name, model,
                          result_class_name, role, result_role, keys_only)

    @cmpi_logging.trace_method
    def cim_method_checkintegrity(self, env, object_name):
        """Implements LMI_SoftwarePackage.CheckIntegrity()

        Verify existence and attributes of files installed from RPM
        package. If all files installed exist and their attributes
        matches, method returns "Pass". "Not passed" is returned when
        arbitrary file differs in its attributes. And "Error" if
        verification could not be done.

        Keyword arguments:
        env -- Provider Environment (pycimmb.ProviderEnvironment)
        object_name -- A pywbem.CIMInstanceName or pywbem.CIMCLassName
            specifying the object on which the method CheckIntegrity()
            should be invoked.

        Returns a two-tuple containing the return value (
            type pywbem.Uint32 SoftwareInstalledPackage.Values.CheckIntegrity)
        and a list of CIMParameter objects representing the output parameters

        Output parameters:
        Failed -- (type REF (pywbem.CIMInstanceName(
                classname='LMI_SoftwareFileCheck', ...))
            Array of references to LMI_SoftwareFileCheck, that did not pass
            verification.

        Possible Errors:
        CIM_ERR_ACCESS_DENIED
        CIM_ERR_INVALID_PARAMETER (including missing, duplicate,
            unrecognized or otherwise incorrect parameters)
        CIM_ERR_NOT_FOUND (the target CIM Class or instance does not
            exist in the specified namespace)
        CIM_ERR_METHOD_NOT_AVAILABLE (the CIM Server is unable to honor
            the invocation request)
        CIM_ERR_FAILED (some other unspecified error occurred)

        """
        if not "Software" in object_name:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing Software property.")
        if not "System" in object_name:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing System property.")

        failed = []
        with YumDB.getInstance() as ydb:
            pkg_info = SoftwarePackage.object_path2pkg(object_name['Software'],
                    kind="installed")
            pkg_check = ydb.check_package(pkg_info)
            for pkg_file in pkg_check.files.values():
                file_check = SoftwareFileCheck.test_file(
                        pkg_check.file_checksum_type, pkg_file)
                if SoftwareFileCheck.filecheck_passed(file_check):
                    continue
                failed.append(SoftwareFileCheck.filecheck2model(
                    pkg_info, pkg_check, pkg_file.path,
                    keys_only=True, file_check=file_check))
        out_params = [ pywbem.CIMParameter('Failed', type='reference',
                value=failed) ]
        return ( getattr(SoftwareInstalledPackage.Values.CheckIntegrity,
                    'Pass' if len(failed) == 0 else 'Not_passed')
               , out_params )

    @cmpi_logging.trace_method
    def cim_method_update(self, env, object_name,
            param_epoch=None,
            param_release=None,
            param_version=None):
        """Implements LMI_SoftwareInstalledPackage.Update()

        Updates the package to latest version. When any of "version" or
        "release" argument is given, install only matching available
        package. Otherwise try to update to newest package available.

        Keyword arguments:
        env -- Provider Environment (pycimmb.ProviderEnvironment)
        object_name -- A pywbem.CIMInstanceName or pywbem.CIMCLassName
            specifying the object on which the method Update()
            should be invoked.
        param_release --  The input parameter release (type unicode)
            Specify particular release of package to update to. Update to
            newest, when empty
        param_version --  The input parameter version (type unicode)
            Specify particular version of package to update to. Update to
            newest, when empty

        Returns a two-tuple containing the return value (
            type pywbem.Uint16 SoftwareInstalledPackage.Values.Update)
        and a list of CIMParameter objects representing the output parameters

        Output parameters:
        Installed -- (type REF (pywbem.CIMInstanceName(
                classname='LMI_SoftwareInstalledPackage', ...))
            The reference to newly installed package, if installation was
            successful. Otherwise reference to current package.

        Possible Errors:
        CIM_ERR_ACCESS_DENIED
        CIM_ERR_INVALID_PARAMETER (including missing, duplicate,
            unrecognized or otherwise incorrect parameters)
        CIM_ERR_NOT_FOUND (the target CIM Class or instance does not
            exist in the specified namespace)
        CIM_ERR_METHOD_NOT_AVAILABLE (the CIM Server is unable to honor
            the invocation request)
        CIM_ERR_FAILED (some other unspecified error occurred)

        """
        if not "Software" in object_name:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing Software property.")
        if not "System" in object_name:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "Missing System property.")
        ComputerSystem.check_path_property(env, object_name, 'System')

        with YumDB.getInstance() as ydb:
            orig = SoftwarePackage.object_path2pkg(object_name['Software'],
                    kind='installed')

            # NOTE: that we need to obtain all available and installed packages
            # because they are disjunctive and we want to select package
            # with highest version. Let the yum do the sorting of packages.
            pkglist = ydb.filter_packages('all',
                    allow_duplicates=True, sort=True,
                    name=orig.name,
                    epoch=param_epoch,
                    version=param_version,
                    release=param_release,
                    arch=orig.arch)

            if len(pkglist) < 1:
                cmpi_logging.logger.error(
                        "desired package matching evr not found")
                raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                        "package matching desired evr not found among"
                        " available packages")
            if len(pkglist) > 1:
                cmpi_logging.logger.info("multiple packages matching"
                        " evr - selecting newest")
                cmpi_logging.logger.debug("matching packages (from oldest):"
                    " [%s]" % ", ".join([str(p) for p in pkglist]))
            desired_pkg = pkglist[-1]

            out_params = [pywbem.CIMParameter('Installed', type='reference')]
            if desired_pkg.installed:
                out_params[0].value = SoftwarePackage.pkg2model(desired_pkg)
                cmpi_logging.logger.info('already up to date')
                return ( SoftwareInstalledPackage.Values.Update.Already_newest
                       , out_params)

            cmpi_logging.logger.info(
                    'trying to update package \"%s\" to "%s"' % (
                        orig, desired_pkg))
            installed_pkg = ydb.update_to_package(desired_pkg)
            cmpi_logging.logger.info('update successful')

            out_params[0].value = SoftwarePackage.pkg2model(installed_pkg)

        return ( SoftwareInstalledPackage.Values.Update.Successful_installation
               , out_params)

